﻿using BenchmarkDotNet.Attributes;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Problems.DataAnalysis.Symbolic.ConstantsOptimization;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
using HeuristicLab.Problems.Instances.DataAnalysis;
using HeuristicLab.Random;
using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Benchmarks {
  [ClrJob]
  [MinColumn, MaxColumn, MeanColumn, MedianColumn]
  public class ConstantsOptimizationPerformance {
    private static readonly int seed = 1234;
    private static readonly int maxTreeSize = 50;
    private static readonly int nTrees = 10;

    private List<ISymbolicExpressionTree> globalTrees = new List<ISymbolicExpressionTree>();
    private List<ISymbolicExpressionTree> iterationTrees = new List<ISymbolicExpressionTree>();
    private IRegressionProblemData problemData;
    private TypeCoherentExpressionGrammar grammar;
    private int[] rows;

    private readonly ISymbolicDataAnalysisExpressionTreeInterpreter interpreter = new SymbolicDataAnalysisExpressionTreeLinearInterpreter();

    [GlobalSetup]
    public void Setup() {
      var twister = new MersenneTwister((uint)seed);
      problemData = new RegressionRealWorldInstanceProvider().LoadData(new Tower());
      rows = Enumerable.Range(0, Rows).ToArray();

      grammar = new TypeCoherentExpressionGrammar();
      grammar.ConfigureAsDefaultRegressionGrammar();

      // generate random trees and only keep trees which are ok
      while (globalTrees.Count < nTrees) {
        var t = Util.CreateRandomTrees(twister, problemData.Dataset, grammar, 1, 1, maxTreeSize, 0, 0).Single();
        Util.InitTree(t, twister, problemData.AllowedInputVariables.ToList());
        var q = SymbolicRegressionSingleObjectivePearsonRSquaredEvaluator.Calculate(interpreter, t, double.MinValue, double.MaxValue, problemData, rows, applyLinearScaling: true);
        if (q > 0) globalTrees.Add(t);
      }
    }

    [IterationSetup]
    public void IterationSetup() {
      iterationTrees.Clear();
      for (int i = 0; i < globalTrees.Count; i++) {
        var t = (ISymbolicExpressionTree)globalTrees[i].Clone(); // trees are changed in OptimizeConstants, so we need to clone the original trees for each iteration
        iterationTrees.Add(t);
      }
    }

    //public IEnumerable<int> RowsValues => Util.IntRange(100, 5000, 100, closed: true);
    //public IEnumerable<int> RowsValues => new[] { 100, 250, 500, 750, 1000, 1500, 2000, 3000, 4000, 4999 };
    public IEnumerable<int> RowsValues => new[] { 4999 };
    public IEnumerable<int> IterationsValues => Util.IntRange(5, 30, 5, closed: true);
    //public IEnumerable<int> IterationsValues => new[] { 0 };

    [ParamsSource(nameof(RowsValues))]
    public int Rows { get; set; }

    [ParamsSource(nameof(IterationsValues))]
    public int Iterations { get; set; }

    [Benchmark]
    public double Tower() {
      double avgQuality = 0;

      if (Iterations > 0) {
        for (int i = 0; i < iterationTrees.Count; i++) {
          avgQuality += LMConstantsOptimizer.OptimizeConstants(iterationTrees[i], problemData.Dataset, problemData.TargetVariable, rows, applyLinearScaling: true, maxIterations: Iterations);
        }
      } else {
        for (int i = 0; i < iterationTrees.Count; i++) {
          avgQuality += SymbolicRegressionSingleObjectivePearsonRSquaredEvaluator.Calculate(interpreter, iterationTrees[i], double.MinValue, double.MaxValue, problemData, rows, applyLinearScaling: true);
        }
      }
      return avgQuality / iterationTrees.Count;
    }
  }
}
